@@ -3,15 +3,17 @@ package DBI::Gofer::Transport::mod_perl;
 use strict;
 use warnings;
 
-our $VERSION = sprintf("0.%06d", q$Revision: 10064 $ =~ /(\d+)/o);
+our $VERSION = 1.015; # keep in sync with Makefile.PL
 
+use UNIVERSAL qw(can);
 use Sys::Hostname qw(hostname);
 use List::Util qw(min max sum);
 
-use DBI qw(dbi_time);
+use DBI 1.605, qw(dbi_time);
 use DBI::Gofer::Execute;
+use Socket;
 
-use constant MP2 => ( exists $ENV{MOD_PERL_API_VERSION} and $ENV{MOD_PERL_API_VERSION} >= 2 );
+use constant MP2 => ( ($ENV{MOD_PERL_API_VERSION}||0) >= 2 or eval "require Apache2::Const");
 BEGIN {
   if (MP2) {
     require Apache2::Connection;
@@ -20,21 +22,21 @@ BEGIN {
     require Apache2::RequestUtil;
     require Apache2::Response;
     require Apache2::Const;
-    Apache2::Const->import(qw(OK SERVER_ERROR));
+    Apache2::Const->import(qw(OK DECLINED SERVER_ERROR));
     require APR::Base64;
     *encode_base64 = \&APR::Base64::encode;
     *decode_base64 = \&APR::Base64::decode;
     *escape_html = sub {
-	my $s = shift;
-	$s =~ s/&/&amp;/g;
-	$s =~ s/</&lt;/g;
-	$s =~ s/>/&gt;/g;
-	return $s;
+        my $s = shift;
+        $s =~ s/&/&amp;/g;
+        $s =~ s/</&lt;/g;
+        $s =~ s/>/&gt;/g;
+        return $s;
     }
   }
   else {
     require Apache::Constants;
-    Apache::Constants->import(qw(OK SERVER_ERROR));
+    Apache::Constants->import(qw(OK DECLINED SERVER_ERROR));
     require Apache::Util;
     Apache::Util->import(qw(escape_html));
     require MIME::Base64;
@@ -42,14 +44,16 @@ BEGIN {
   }
 }
 
+use DBI::Gofer::Serializer::DataDumper;
+
 use base qw(DBI::Gofer::Transport::Base);
 
-my $transport = __PACKAGE__->new();
+our $transport = __PACKAGE__->new();
 
-my %executor_configs = ( default => { } );
-my %executor_cache;
-
-my $datadumper_serializer = DBI::Gofer::Serializer::DataDumper->new;
+our %executor_configs = ( default => { } );
+our %executor_cache;
+our $show_client_hostname_in_status = 1;
+our $datadumper_serializer = DBI::Gofer::Serializer::DataDumper->new;
 
 _install_apache_status_menu_items(
     DBI_gofer => [ 'DBI Gofer', \&_apache_status_dbi_gofer ],
@@ -59,64 +63,145 @@ _install_apache_status_menu_items(
 sub handler : method {
     my $self = shift;
     my $r = shift;
+    my $time_received = dbi_time();
     my $headers_in = $r->headers_in;
 
+    my ($frozen_request,  $request,  $request_serializer);
+    my ($frozen_response, $response, $response_serializer);
+    my $executor;
+
+    my $http_status = SERVER_ERROR;
+    my $remote_ip = $headers_in->{Client_ip}    # e.g., cisco load balancer
+        || $headers_in->{'X-Forwarded-For'}     # e.g., mod_proxy (XXX may contain more than one ip)
+        || $r->connection->remote_ip;
+
     eval {
-        my $time_received = dbi_time();
-        my $executor = $self->executor_for_apache_request($r);
+        $executor = $self->executor_for_apache_request($r);
 
         my $request_content_length = $headers_in->{'Content-Length'};
-	my $frozen_request;
+        # XXX get content-type by response_content_type() meth call on serializer?
+        # (need to think-through content-type, transfer-encoding, disposition etc etc
         my $response_content_type = 'application/x-perl-gofer-response-binary';
-	my $response_serializer;
-	# should probably contol flow via method: GET vs POST
-	my $of = "";
-	if (!$request_content_length) {
-	    my $args = $r->args || '';
-	    my %args = map { (split('=',$_,2))[0,1] } split /[&;]/, $args, -1;
-	    my $req = $args{req} or die "No req argument or Content-Length ($args)\n";
-	    $frozen_request = decode_base64($req);
-	    if ($args{_dd}) { # temp hack
-	    $response_serializer = $datadumper_serializer;
-	    $response_content_type = 'text/plain';
-	    }
-	}
-	else {
-	    my $content_type = $headers_in->{'Content-Type'};
-	    die "Unsupported gofer Content-Type"
-		unless $content_type eq 'application/x-perl-gofer-request-binary';
-	    $r->read($frozen_request, $request_content_length);
-	}
-        my $request = $transport->thaw_request($frozen_request);
+        # XXX should probably contol flow via method: GET vs POST
+        my $of = "";
+        if (!$request_content_length) { # assume GET request
+            my $args = $r->args || '';
+            my %args = map { (split('=',$_,2))[0,1] } split /[&;]/, $args, -1;
+            my $req = $args{req}
+                or die "No req argument or Content-Length ($args)\n";
+            $frozen_request = decode_base64($req);
 
-        my $response = $executor->execute_request( $request );
+            if ($args{_dd}) { # XXX temp hack
+                $response_serializer = $datadumper_serializer;
+                $response_content_type = 'text/plain';
+                if ($args{_dd} eq 'request') { # XXX even more of a temp hack
+                    $request = $transport->thaw_request($frozen_request);
+                    $r->pnotes(gofer_request => $request);
+                    $frozen_response = $datadumper_serializer->serialize($request);
+                    goto send_frozen_response;
+                }
+            }
+        }
+        else {
+            my $content_type = $headers_in->{'Content-Type'};
+            die "Unsupported gofer Content-Type"
+                unless $content_type eq 'application/x-perl-gofer-request-binary';
+            $r->read($frozen_request, $request_content_length);
+            if (length($frozen_request) != $request_content_length) {
+                die sprintf "Gofer request length (%d) doesn't match Content-Length header (%d)",
+                    length($frozen_request), $request_content_length;
+            }
+        }
 
-        my $frozen_response = $transport->freeze_response($response, $response_serializer);
+        $request = $transport->thaw_request($frozen_request);
+        $r->pnotes(gofer_request => $request);
 
+        $response = $executor->execute_request( $request );
+        $r->pnotes(gofer_response => $response);
+
+        $frozen_response = $transport->freeze_response($response, $response_serializer);
+
+    send_frozen_response:
         $r->content_type($response_content_type);
         # setup http headers
         # See http://perl.apache.org/docs/general/correct_headers/correct_headers.html
         # provide Content-Length for KeepAlive so it works if people want it
-        $r->headers_out->{'Content-Length'} = length($frozen_response);
+        $r->headers_out->{'Content-Length'} = do { use bytes; length($frozen_response) };
 
         $r->print($frozen_response);
 
-        # XXX add as cleanup handler
-        $executor->update_stats($request, $response, $frozen_request, $frozen_response, $time_received);
+        $http_status = OK;
     };
     if ($@) {
         # for errors at this level we don't send a serialized Gofer Response 
-        # just a plain error message and SERVER_ERROR (500) status
-        chomp(my $error = $@);
-        $error .= sprintf " in %s request from %s",
-		$headers_in->{'Content-Type'}||'', $r->connection->remote_ip;
-        warn "$error\n";
-        $r->custom_response(SERVER_ERROR, sprintf "%s. (%s %s, DBI %s, on %s)",
-            $error, __PACKAGE__, $VERSION, $DBI::VERSION, hostname());
-        return SERVER_ERROR;
+        # (but we do create one for logging/stats purposes)
+
+        $http_status = SERVER_ERROR;
+
+        # discard any response that might have been prepared already
+        # (e.g., an exception is thrown after execute_request returns)
+        $response = undef;
+
+        my $error = $@;
+        my $action;
+        if (ref $error) {
+            # allow the exception to override some things
+            $http_status = $error->{http_status} if $error->{http_status};
+            $action      = $error->{http_action} if $error->{http_action};
+            $response    = $error->{gofer_response} if $error->{gofer_response};
+            $error       = $error->{error_text}  if $error->{error_text};
+            $error       = $error->text if can($error, 'text');
+        }
+
+        chomp $error;
+        $error .= sprintf " in %s request from %s, http status %d",
+                $headers_in->{'Content-Type'}||$r->method, $remote_ip, $http_status;
+
+        # record the error (via cleanup handler below) so we can see it later
+        # remotely if track_recent is enabled.
+        # if exception didn't include a response for logging then create one
+        $response ||= $executor->new_response_with_err($DBI::stderr||1, $error);
+        $r->pnotes(gofer_response => $response);
+        $frozen_response = $transport->freeze_response($response);
+
+        my $default_action = sub {   # default error response behaviour
+            my ($r, $errstr, $http_status) = @_;
+            warn "$errstr\n";
+            $r->status($http_status);
+            $r->content_type("text/plain");
+            $r->custom_response($http_status, sprintf "%s. (%s %s, DBI %s, on %s pid $$)",
+                $errstr, __PACKAGE__, $VERSION, $DBI::VERSION, hostname());
+            return $http_status;
+        };
+        $action ||= $default_action;
+
+        $http_status = $action->($r, $error, $http_status, $default_action);
     }
 
-    return OK;
+    # defer stats until the cleanup phase
+    $r->push_handlers('PerlCleanupHandler', sub {
+        $executor->update_stats(
+            $request,   # may not be defined if error thawing
+            $response,  # always present
+            # if we've used a non-default serializer (ie Data::Dumper)
+            # then don't store the frozen items because we may not
+            # be able to thaw it. XXX needs better approach
+            ($request_serializer ) ? undef : $frozen_request,
+            ($response_serializer) ? undef : $frozen_response,
+            $time_received,
+            {
+                from => $remote_ip,
+            },
+            {
+                r => $r,
+                transport => $transport,
+            },
+        ) if $executor;
+
+        return DECLINED;
+    });
+
+    return $http_status;
 }
 
 
@@ -143,9 +228,6 @@ sub _merge_named_configurations {
 
     push @location_configs, 'default' unless @location_configs;
 
-    # XXX chicken-and-egg - need to move this into loop to get class from individual configs
-    my $gofer_execute_class ||= 'DBI::Gofer::Execute';
-    my $proto_config = $gofer_execute_class->valid_configuration_attributes();
 
     # merge all configs for this location in sequence, later override earlier
     my %merged_config;
@@ -156,6 +238,8 @@ sub _merge_named_configurations {
             # (don't die for 'default' unless it was explicitly requested)
             die "$tag: GoferConfig '$config_name' not defined";
         }
+        my $gofer_execute_class = $config->{gofer_execute_class} || 'DBI::Gofer::Execute';
+        my $proto_config = $gofer_execute_class->valid_configuration_attributes();
         my @info;
         while ( my ($item_name, $proto_type) = each %$proto_config ) {
             next if not exists $config->{$item_name};
@@ -166,8 +250,12 @@ sub _merge_named_configurations {
                 $merged->{$_} = $item_value->{$_} for keys %$item_value;
             }
             else {
-                push @info, "$item_name=$item_value" if $verbose && defined $item_value;
                 $merged_config{$item_name} = $item_value;
+                if ($verbose && exists $config->{$item_name}) {
+                    my $v = $item_value;
+                    $v =~ s/\(0x\w+\)$// if ref $v;
+                    push @info, "$item_name=$v";
+                }
             }
         }
         warn "$tag: GoferConfig $config_name: @info\n" if @info;
@@ -197,7 +285,10 @@ sub _install_apache_status_menu_items {
     my %apache_status_menu_items = @_;
     my $apache_status_class;
     if (MP2) {
-        $apache_status_class = "Apache2::Status" if Apache2::Module::loaded('Apache2::Status');
+        $apache_status_class = "Apache2::Status" if eval {
+            require Apache2::Module;
+            Apache2::Module::loaded('Apache2::Status');
+        };
     }
     elsif ($INC{'Apache.pm'}                       # is Apache.pm loaded?
         and Apache->can('module')               # really?
@@ -226,7 +317,7 @@ sub _apache_status_dbi_gofer {
     local $Data::Dumper::Purity    = 0;
 
     my @s = ("<pre>",
-        "<b>DBI::Gofer::Transport::mod_perl $VERSION</b><p>",
+        "<b>DBI::Gofer::Transport::mod_perl $VERSION</b> - <b>DBI $DBI::VERSION</b><p>",
     );
     my $time_now = dbi_time();
 
@@ -237,95 +328,165 @@ sub _apache_status_dbi_gofer {
     $path_info =~ s!^/perl-status!!;
 
     # hack to enable simple actions to be invoked via the status interface
-    my $action = ($path_info =~ s/:(\w+)$//) ? $1 : undef;
+    # format "...:foo" or "...:foo,opt1=bar,opt2=baz"
+    my $action = ($path_info =~ s/:([\w,=]+)$//) ? $1 : '';
 
     if ($path_info) {
 
         my $executor = $executor_cache{$path_info}
             or return [ "No Gofer executor found for '$path_info'" ];
 
+        ($action, my @actions) = split /,/, $action;
+        my %action_opts = map { split /=/, $_, 2 } @actions;
+
         my $stats = $executor->{stats} ||= {};
-        my $recent_requests = $stats->{recent_requests};
+        my $queue_name = "recent_". ($action_opts{recent}||'requests');
+        my $queue = $stats->{$queue_name};
+        return [ "No $queue_name found for '$path_info'" ]
+            unless ref $queue eq 'ARRAY';
 
         if ($action) {
-		# change to hash of code refs and add links to the actions into the output
+                # change to hash of code refs and add links to the actions into the output
             if ($action eq 'reset_stats') {
                 $executor->{stats} = { _reset_stats_at => scalar localtime(time) };
                 $stats = {};
             }
             elsif ($action eq 'recent_as_urls') {
-		my $host = $r->get_server_name;
-		my $port = $r->get_server_port;
-		@s = ();
-		for my $rr (@$recent_requests) {
-		    my $b64_request = encode_base64($rr->{request});
-		    push @s, "http://$host:$port$path_info?req=$b64_request\n";
-		}
-		return \@s;
-	    }
+                my $host = $r->get_server_name;
+                my $port = $r->get_server_port;
+                @s = ();
+                for my $rr (@$queue) {
+                    my $b64_request = encode_base64($rr->{request});
+                    push @s, "http://$host:$port$path_info?req=$b64_request\n";
+                }
+                return \@s;
+            }
+            elsif ($action eq 'view') {
+                # fall through
+            }
             else {
                 return [ "Unknown action '$action' ignored for $path_info" ];
             }
         }
 
-        # don't Data::Dumper all the recent_requests
+        # don't Data::Dumper all the recent_requests & recent_errors
         local $stats->{recent_requests} = @{$stats->{recent_requests}||[]};
+        local $stats->{recent_errors}   = @{$stats->{recent_errors}  ||[]};
         push @s, escape_html( Data::Dumper::Dumper($executor) );
         push @s, "<hr>";
 
         my ($idle_total, $dur_total, $time_received_prev, $duration_prev) = (0,0,0,0);
-	my @redo_urls;
-        for my $rr (@$recent_requests) {
+        my @redo_urls;
+        my (%from, %dup_reqs);
+        for my $rr (@$queue) {
             my $time_received = $rr->{time_received};
             my $duration = $rr->{duration};
             my $idle = ($time_received_prev) ? abs($time_received-$time_received_prev)-$duration_prev : 0;
             $rr->{_time_received} ||= localtime($time_received);
+            my $from = $rr->{meta}{from};
+            $from{ $from }{requests}++ if $from;
 
             # mark idle periods - handy when testing
             push @s, "<hr>" if $time_received_prev and $idle > 10;
 
-	    my $b64_request = encode_base64($rr->{request});
-	    push @redo_urls, "$path_info?req=$b64_request";
-	    push @s, sprintf qq{\tredo: <a href="%s?req=%s">raw</a>, <a href="%s?_dd=1&req=%s">dump</a>},
-		$path_info, $b64_request,
-		$path_info, $b64_request;
+            if (my $request = $rr->{request}) {
+                my $b64_request = encode_base64($rr->{request});
+                push @redo_urls, "$path_info?req=$b64_request";
+                push @s, sprintf qq{\tredo: <a href="%s?req=%s">raw</a>, <a href="%s?_dd=1&req=%s">dump</a>},
+                    $path_info, $b64_request,
+                    $path_info, $b64_request;
 
-            my $request  = $transport->thaw_request($rr->{request});
-            push @s, escape_html( $request->summary_as_text({
-                at => $rr->{_time_received},
-                age => int($time_now-$time_received),
-                idle => $idle,
-                size => length($rr->{request}),
-            }) );
+                my $is_dupreq = $dup_reqs{ $rr->{request} }++;
 
-            my $response = $transport->thaw_response($rr->{response});
-            push @s, escape_html( $response->summary_as_text({
-                duration => $duration,
-                size => length($rr->{response}),
-            }) );
+                my $request_html = eval {
+                    my $request  = $rr->{request_object}
+                        || $transport->thaw_request($rr->{request});
+                    escape_html( $request->summary_as_text({
+                        at => $rr->{_time_received},
+                        age => int($time_now-$time_received),
+                        idle => $idle,
+                        size => length($rr->{request}),
+                        ($from) ? (from => $from) : (),
+                        ($is_dupreq) ? (is_dup => $is_dupreq) : (),
+                    }) );
+                } || escape_html("ERROR THAWING REQUEST: $@");
 
-            push @s, "";
+                # bold the word _dup if it's a dup
+                $request_html =~ s{\b is_dup \b}{<b>is_dup</b>}x if $is_dupreq;
+                push @s, $request_html;
+            }
+            else {
+                push @s, "<i>(no request data)</i>\n";
+            }
+
+            my $response_html = eval {
+                my $response = $rr->{response_object}
+                    || $transport->thaw_response($rr->{response});
+                $from{ $from }{errors}++ if $from && $response->err;
+                escape_html( $response->summary_as_text({
+                    duration => $duration,
+                    ($rr->{response}) ? (size => length $rr->{response}) : (),
+                }) );
+            } || escape_html("ERROR THAWING RESPONSE: $@");
+            push @s, $response_html;
+
+            push @s, "\n";
 
             $idle_total += $idle;
             $dur_total  += $duration;
             ($time_received_prev, $duration_prev) = ($time_received, $duration);
         }
         push @s, "<hr>\n";
-        if (@$recent_requests) {
-            my @rr_requ_size = map { length($_->{request}) }  @$recent_requests;
-            push @s, sprintf "Recent request size:  min %d, avg %d, max %d (sum %d for last %d)\n",
-                min(@rr_requ_size), sum(@rr_requ_size)/@rr_requ_size, max(@rr_requ_size), sum(@rr_requ_size), scalar @rr_requ_size;
+        if (@$queue) {
+            my $time_span = $dur_total+$idle_total;
+            push @s, sprintf "Summary for the %d requests shown above (covering %d seconds for pid $$)...\n",
+                scalar @$queue, $time_span;
 
-            my @rr_resp_size = map { length($_->{response}) } @$recent_requests;
-            push @s, sprintf "Recent response size: min %d, avg %d, max %d (sum %d for last %d)\n",
-                min(@rr_resp_size), sum(@rr_resp_size)/@rr_resp_size, max(@rr_resp_size), sum(@rr_resp_size), scalar @rr_resp_size;
+            my @rr_requ_size = map { length($_->{request}||'') }  @$queue;
+            push @s, sprintf "Request  size: min %4d, avg %4d, max %4d (sum %d \@ %dB/sec)\n",
+                min(@rr_requ_size), sum(@rr_requ_size)/@rr_requ_size, max(@rr_requ_size),
+                sum(@rr_requ_size), sum(@rr_requ_size)/$time_span;
 
-            my @rr_resp_dur = map { $_->{duration} } @$recent_requests;
-            push @s, sprintf "Recent response time: min %.3fs, avg %.3fs, max %.3fs (sum %d for last %d)\n",
-                min(@rr_resp_dur), sum(@rr_resp_dur)/@rr_resp_dur, max(@rr_resp_dur), sum(@rr_resp_dur), scalar @rr_resp_dur;
+            my @rr_resp_size = map { length($_->{response}||'') } @$queue;
+            push @s, sprintf "Response size: min %4d, avg %4d, max %4d (sum %d \@ %dB/sec)\n",
+                min(@rr_resp_size), sum(@rr_resp_size)/@rr_resp_size, max(@rr_resp_size),
+                sum(@rr_resp_size), sum(@rr_resp_size)/$time_span;
 
-            push @s, sprintf "Occupancy for those %d requests: %.1f%% (%.3fs busy, %.3fs idle)\n",
-                scalar @$recent_requests, $dur_total/($dur_total+$idle_total)*100, $dur_total, $idle_total
+            my @rr_resp_dur = map { $_->{duration} } @$queue;
+            push @s, sprintf "Response time: min %.3fs, avg %.3fs, max %.3fs\n",
+                min(@rr_resp_dur), sum(@rr_resp_dur)/@rr_resp_dur, max(@rr_resp_dur), sum(@rr_resp_dur);
+
+            push @s, sprintf "Request rate: %.1f/min (occupancy: %.1f%% with %.3fs busy and %.3fs idle)\n",
+                    @$queue/($time_span/60),
+                    $dur_total/($dur_total+$idle_total)*100, $dur_total, $idle_total
+                if $queue_name eq 'recent_requests';
+
+            if ( my @dups = grep { $_ > 1 } values %dup_reqs ) {
+                push @s, sprintf "Duplicate requests: %d distinct duplicates, total %d duplicates\n",
+                    scalar @dups, sum(@dups);
+            }
+
+            if ($show_client_hostname_in_status) { # use DNS lookup
+                eval {
+                    local $SIG{ALRM} = "TIMEOUT DNS ".__PACKAGE__;
+                    alarm(5);
+                    for my $from (keys %from) {
+                        next unless $from =~ /^\d+\./;
+                        my $new = sprintf "%s %s",
+                            gethostbyaddr(inet_aton($from),AF_INET) || "?",
+                            $from;
+                        $from{ $new } = delete $from{ $from };
+                    }
+                    alarm(0);
+                };
+                alarm(0);
+                warn $@ if $@;
+            }
+            push @s, sprintf "Recent request distribution from %d sources:\n", scalar keys(%from)
+                if keys(%from);
+            push @s, sprintf "%-20s: %3d, errors %d\n",
+                    $_, $from{$_}{requests}, $from{$_}{errors}||0
+                for sort keys %from;
         }
         return \@s;
     }
@@ -514,3 +675,5 @@ modify it under the same terms as Perl itself. See L<perlartistic>.
 L<DBD::Gofer> and L<DBD::Gofer::Transport::http>.
 
 =cut
+
+# vim: ts=8:sw=4:et
